from __future__ import annotations

import sys
from io import BytesIO, TextIOWrapper
import typing as t

import click
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QObject, QEvent

from clickqt.core.error import ClickQtError
from clickqt.widgets.basewidget import BaseWidget
from clickqt.widgets.nvaluewidget import NValueWidget


class FocusOutValidator(QWidget):
    """Validates a widget value when the widget goes out of focus.

    :param widget: The clickqt widget that will be watched for losing focus
    """

    def __init__(self, widget: BaseWidget):
        super().__init__()

        self.widget = widget

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        """Inherited from :class:`~PySide6.QtWidgets.QWidget`\n
        If the widget went out of focus, its value will be validated.
        Print-statements that may occur in user-defined callbacks will not be printed.
        """

        if event.type() == QEvent.Type.FocusOut:
            # Don't print callback prints
            old_strerr = sys.stderr
            old_strout = sys.stdout
            sys.stderr = sys.stdout = TextIOWrapper(BytesIO(), "utf-8")

            # Don't set the new value because the callback call could reject the (new) value
            # when trying to execute the command (=user clicked on the "Run"-button)
            self.__validate(self.widget)

            sys.stderr = old_strerr
            sys.stdout = old_strout

        return QWidget.eventFilter(self, watched, event)

    def __validate(self, widget: BaseWidget) -> tuple[t.Any, ClickQtError]:
        """Validates the value of the widget that went out of focus."""

        if widget.parent_widget is not None and not isinstance(
            widget.parent_widget, NValueWidget
        ):
            return self.__val(widget)
        if widget.parent_widget is None:
            return widget.get_value()

        # self.widget.parent_widget == NValueWidget -> We have a child here

        try:  # Try to convert the provided value into the corresponding click object type
            ret_val = widget.type.convert(
                value=widget.get_widget_value(),
                param=widget.param,
                ctx=click.Context(widget.click_command),
            )
            # Don't consider callbacks because we have only one child here
            widget.handle_valid(True)
            return (ret_val, ClickQtError())
        except Exception as e: # pylint: disable=broad-exception-caught
            widget.handle_valid(False)
            return (None, ClickQtError(ClickQtError.ErrorType.CONVERTING_ERROR, widget.widget_name, e))

    def __val(self, widget: BaseWidget) -> tuple[t.Any, ClickQtError]:
        """Calls get_value() on the widget with no parent."""

        if widget.parent_widget is not None:
            return self.__val(widget.parent_widget)
        return widget.get_value()



from __future__ import annotations

import typing as t

import click
from PySide6.QtWidgets import QGroupBox, QHBoxLayout

from clickqt.widgets.basewidget import BaseWidget, MultiWidget


class TupleWidget(MultiWidget):
    """Represents a click.types.Tuple-object.
    The child widgets are set according to :func:`~clickqt.widgets.basewidget.MultiWidget.init`.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param widgetsource: A reference to :func:`~clickqt.core.gui.GUI.create_widget`
    :param parent: The parent BaseWidget of **otype**, defaults to None. Needed for :class:`~clickqt.widgets.basewidget.MultiWidget`-widgets
    :param kwargs: Additionally parameters ('com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QGroupBox  #: The Qt-type of this widget.

    def __init__(
        self,
        otype: click.ParamType,
        param: click.Parameter,
        widgetsource: t.Callable[[t.Any], BaseWidget],
        parent: t.Optional[BaseWidget] = None,
        **kwargs,
    ):
        super().__init__(otype, param, parent=parent, **kwargs)

        assert isinstance(
            otype, click.Tuple
        ), f"'otype' must be of type '{click.Tuple}', but is '{type(otype)}'."
        assert otype.is_composite, "otype.is_composite should be True"

        self.widget.setLayout(QHBoxLayout())

        for child_type in otype.types if hasattr(otype, "types") else otype:
            nargs = self.param.nargs
            self.param.nargs = 1
            bw: BaseWidget = widgetsource(
                child_type, self.param, widgetsource=widgetsource, parent=self, **kwargs
            )
            self.param.nargs = nargs
            bw.layout.removeWidget(bw.label)
            bw.label.deleteLater()
            self.widget.layout().addWidget(bw.container)
            self.children.append(bw)

        self.init()



from __future__ import annotations

import typing as t

import click
from PySide6.QtWidgets import QWidget, QHBoxLayout

from clickqt.widgets.basewidget import BaseWidget
from clickqt.core.error import ClickQtError


class ConfirmationWidget(BaseWidget):
    """Represents a click option with confirmation_prompt==True. It stores two clickqt widgets of the same type.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param widgetsource: A reference to :func:`~clickqt.core.gui.GUI.create_widget`
    :param kwargs: Additionally parameters ('parent', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QWidget  #: The Qt-type of this widget. It's a container for storing the two input-widgets

    def __init__(
        self,
        otype: click.ParamType,
        param: click.Parameter,
        widgetsource: t.Callable[[t.Any], BaseWidget],
        **kwargs
    ):
        super().__init__(otype, param, **kwargs)

        assert (
            hasattr(self.param, "confirmation_prompt")
            and self.param.confirmation_prompt
        ), "'param.confirmation_prompt' should be True"

        self.param.confirmation_prompt = False  # Stop recursion
        self.field: BaseWidget = widgetsource(
            self.type, param, parent=self, vboxlayout=True, **kwargs
        )  #: First input widget.
        kwargs["label"] = "Confirmation "
        self.confirmation_field: BaseWidget = widgetsource(
            self.type, param, parent=self, vboxlayout=True, **kwargs
        )  #: Second (confirmation) input widget.
        self.param.confirmation_prompt = True
        self.widget.setLayout(QHBoxLayout())
        self.layout.removeWidget(self.label)
        self.layout.removeWidget(self.widget)
        self.label.deleteLater()
        self.container.deleteLater()
        self.container = self.widget
        self.container.layout().setContentsMargins(0, 0, 0, 0)
        self.widget.layout().addWidget(self.field.container)
        self.widget.layout().addWidget(self.confirmation_field.container)

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, None))

    def set_value(self, value: t.Any):
        """Sets **value** as widget value for :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.field` and
        :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.confirmation_field` according to :func:`~clickqt.widgets.basewidget.BaseWidget.set_value`."""

        if value is not None:
            self.field.set_value(value)
            self.confirmation_field.set_value(value)

    def handle_valid(self, valid: bool):
        """Changes the widget border for :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.field` and
        :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.confirmation_field` according to :func:`~clickqt.widgets.basewidget.BaseWidget.handle_valid`."""

        self.field.handle_valid(valid)
        self.confirmation_field.handle_valid(valid)

    def get_value(self) -> tuple[t.Any, ClickQtError]:
        """Calls :func:`~clickqt.widgets.basewidget.BaseWidget.get_value` on :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.field` and
        :attr:`~clickqt.widgets.confirmationwidget.ConfirmationWidget.confirmation_field`, validates the result and returns it.

        :return: Valid: (widget value or the value of a callback, :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
                 Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.CONVERTING_ERROR` or
                 :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR` or :class:`~clickqt.core.error.ClickQtError.ErrorType.CONFIRMATION_INPUT_NOT_EQUAL_ERROR`)
        """

        val1, err1 = self.field.get_value()
        val2, err2 = self.confirmation_field.get_value()

        if (
            err1.type != ClickQtError.ErrorType.NO_ERROR
            or err2.type != ClickQtError.ErrorType.NO_ERROR
        ):
            return (
                None,
                err1 if err1.type != ClickQtError.ErrorType.NO_ERROR else err2,
            )

        if val1 != val2:
            self.handle_valid(False)
            return (
                None,
                ClickQtError(
                    ClickQtError.ErrorType.CONFIRMATION_INPUT_NOT_EQUAL_ERROR,
                    self.widget_name,
                ),
            )
        self.handle_valid(True)
        return (val1, ClickQtError())

    def get_widget_value(self) -> t.Any:
        return self.field.get_widget_value()



import click
import clickqt
import os
import re


def locate_lidar_dataset(ctx, param, path):
    """Expand given data paths into a list of files"""

    # Validate that the file has the las or laz extension
    _, ext = os.path.splitext(path)
    if ext.lower() in (".las", ".laz"):
        return path
    else:
        raise click.BadParameter(f"Lidar datasets must be .las or .laz (not: {path})")


def validate_segmentation(ctx, param, filename):
    # Check that the file has the correct extension
    _, ext = os.path.splitext(filename)
    if ext.lower() != ".geojson":
        raise click.BadParameter(
            f"Segmentation files must be .geojson (not: {filename})"
        )

    # Try reading the content
    try:
        seg = filename
    except:
        # TODO: Restrict the exception we are expecting here
        raise click.BadParameter(f"Segmentation file was not {filename} readable")

    return seg


def validate_suffix(ctx, param, suffix):
    if not re.fullmatch("[a-z0-9_]*", suffix):
        raise click.BadParameter(
            f"Suffix should consist of lowercase letters, numbers and underscores only"
        )
    return suffix


def validate_spatial_reference(ctx, param, crs):
    try:
        return crs
    except:
        raise click.BadParameter(
            f"Cannot validate spatial reference system '{crs}'. Use either WKT or 'EPSG:xxxx'"
        )


@click.command()
@click.option(
    "--dataset",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    callback=locate_lidar_dataset,
    help="The LAS/LAZ data file to work on.",
)
@click.option(
    "--dataset-crs",
    type=str,
    required=True,
    callback=validate_spatial_reference,
    help="The CRS of the data",
)
@click.option(
    "--segmentation",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    callback=validate_segmentation,
    help="The GeoJSON file that describes the segmentation of the dataset. This is expected to be generated either by the Jupyter UI or otherwise provide the necessary information about what filter pipelines to apply.",
)
@click.option(
    "--segmentation-crs",
    type=str,
    required=True,
    callback=validate_spatial_reference,
    help="The CRS used in the segmentation",
)
@click.option(
    "--library",
    type=click.Path(exists=True, file_okay=False),
    multiple=True,
    help="A filter library location that AFwizard should take into account. Can be given multiple times.",
)
@click.option(
    "--output-dir",
    type=click.Path(file_okay=False),
    default="output",
    help="The directory to place output files (both LAS/LAZ and GeoTiff).",
    show_default=True,
)
@click.option(
    "--resolution",
    type=click.FloatRange(min=0.0, min_open=True),
    default=0.5,
    help="The meshing resolution to use for generating GeoTiff files",
    metavar="FLOAT",
    show_default=True,
)
@click.option(
    "--compress",
    type=bool,
    is_flag=True,
    help="Whether LAZ files should be written instead of LAS.",
)
@click.option(
    "--suffix",
    type=str,
    default="filtered",
    help="The suffix to add to filtered datasets.",
    callback=validate_suffix,
    show_default=True,
)
@click.option(
    "--opals-dir",
    type=click.Path(file_okay=False, exists=True),
    help="The directory where to find an OPALS installation",
)
@click.option(
    "--lastools-dir",
    type=click.Path(file_okay=False, exists=False),
    help="The directory where to find a LASTools installation",
)
def main(**args):
    print("main executed")


gui = clickqt.qtgui_from_click(main)

if __name__ == "__main__":
    gui()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Project information -----------------------------------------------------

project = "clickqt"
copyright = "2022, Dominic Kempf"
author = "Dominic Kempf"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_mdinclude",
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
    "enum_tools.autoenum",
    "sphinx_qt_documentation",
]

# Make the documentations order the same as the source order.
autodoc_member_order = "bysource"

# Linking to PySide6 documentation on "https://doc.qt.io/qtforpython-6/"
qt_documentation = "PySide6"

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# Copy readme_resources images into documentation so they show up when
# markdown files with relative paths are transcluded
import os
import shutil


def copy_readme_resources(app, docname=None):
    if app.builder.name == "html":
        output_dir = os.path.join(app.outdir, "readme_resources")
        source_dir = os.path.join(app.srcdir, "..", "readme_resources")
        if not os.path.exists(output_dir):
            shutil.copytree(source_dir, output_dir)


def setup(app):
    app.connect("builder-inited", copy_readme_resources)



from __future__ import annotations

import typing as t
import click
from PySide6.QtWidgets import QMessageBox, QWidget
from clickqt.core.error import ClickQtError
from clickqt.widgets.basewidget import BaseWidget


class MessageBox(BaseWidget):
    """Represents clicks Yes-Parameter that asks a user for confirmation. This widget is a confirmation dialog which does not
    appear in the GUI, but pops up one step before execution the command.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QWidget  #: The Qt-type of this widget. Its just a container for storing the messagebox.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert (
            hasattr(param, "is_flag") and param.is_flag
        ), "'param.is_flag' should be True"
        assert (
            hasattr(param, "prompt") and param.prompt
        ), "'param.prompt' should be not empty"

        self.yes: bool = False  #: Confirmation of the user, defaults to False

        if self.parent_widget is None:
            self.set_value(BaseWidget.get_param_default(param, False))

        self.layout.removeWidget(self.label)
        self.layout.removeWidget(self.widget)
        self.label.deleteLater()
        self.container.deleteLater()
        self.container = self.widget
        self.container.setVisible(False)

    def set_value(self, value: t.Any):
        self.yes = bool(
            self.type.convert(
                str(value), self.click_command, click.Context(self.click_command)
            )
        )

    def get_value(self) -> tuple[t.Any, ClickQtError]:
        """Opens a confirmation dialog, passes the decision to :func:`~clickqt.widgets.basewidget.BaseWidget.get_value`
        and returns the result. See also :func:`~clickqt.widgets.messagebox.MessageBox.get_widget_value`

        :return: Valid: (widget value or the value of a callback, :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
                 Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.CONVERTING_ERROR` or
                 :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR`)
        """

        if (
            QMessageBox.information(
                self.widget,
                "Confirmation",
                str(self.param.prompt),
                QMessageBox.Yes | QMessageBox.No,
            )
            == QMessageBox.Yes
        ):
            self.yes = True
        else:
            self.yes = False

        return super().get_value()

    def get_widget_value(self) -> bool:
        """Returns the user-decision of the confirmation dialog.

        :return: True: The user clicked on the "Yes"-button\n
                 False: The user clicked on the "No"-button
        """

        return self.yes



from .basewidget import BaseWidget, NumericField, ComboBoxBase, MultiWidget
from .checkbox import CheckBox
from .textfield import TextField, PathField
from .combobox import ComboBox, CheckableComboBox
from .datetimeedit import DateTimeEdit
from .numericfields import IntField, RealField
from .passwordfield import PasswordField
from .filepathfield import FilePathField
from .filefield import FileField
from .confirmationwidget import ConfirmationWidget
from .messagebox import MessageBox
from .tuplewidget import TupleWidget
from .nvaluewidget import NValueWidget
from .multivaluewidget import MultiValueWidget



""" Contains the FileField class """
from __future__ import annotations

import sys
from io import StringIO, BytesIO
import typing as t

import click
from PySide6.QtWidgets import QLineEdit, QInputDialog

from clickqt.core.error import ClickQtError
from clickqt.widgets.textfield import PathField


class FileField(PathField):
    """Represents a click.types.File object.

    :param otype: The type which specifies the clickqt widget type.
        This type may differ from **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additional parameters ('parent', 'widgetsource', 'com', 'label') needed for
        :class:`~clickqt.widgets.basewidget.MultiWidget`- /
        :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`- widgets
    """

    widget_type = QLineEdit  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, click.File
        ), f"'otype' must be of type '{click.File}', but is '{type(otype)}'."

        self.file_type: PathField.FileType = (
            PathField.FileType.File
        )  #: File type is a :attr:`~clickqt.widgets.textfield.PathField.FileType.File`.

    def get_value(self) -> tuple[t.Any, ClickQtError]:
        """
        Opens an input dialogue that represents sys.stdin if 'r' is in **otype**.mode
            and the current widget value is '-', passes the input to
        :func:`~clickqt.widgets.basewidget.BaseWidget.get_value`, and returns the result.

        :return:
            Valid: (widget value or the value of a callback,
                :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
            Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.CONVERTING_ERROR` or
                :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR` or
                :class:`~clickqt.core.error.ClickQtError.ErrorType.ABORTED_ERROR`)
        """

        if "r" in self.type.mode and self.widget.text() == "-":
            self.handle_valid(True)

            def ret():  # FocusOutValidator should not open this dialog
                user_input, is_ok = QInputDialog.getMultiLineText(
                    self.widget, "Stdin Input", self.label.text()
                )
                if not is_ok:
                    return (None, ClickQtError(ClickQtError.ErrorType.ABORTED_ERROR))

                old_stdin = sys.stdin
                sys.stdin = (
                    BytesIO(user_input.encode(sys.stdin.encoding))
                    if "b" in self.type.mode
                    else StringIO(user_input)
                )
                val = super(  # pylint: disable=super-with-arguments
                    FileField, self
                ).get_value()
                sys.stdin = old_stdin
                return val

            return (ret, ClickQtError())
        return super().get_value()



from __future__ import annotations

import typing as t

import pytest
import click
from PySide6.QtWidgets import QTabWidget, QPushButton, QSplitter, QWidget
from PySide6.QtCore import Qt, QThread

import clickqt
from tests.testutils import ClickAttrs, raise_, wait_process_Events
from clickqt.core.output import TerminalOutput
from clickqt.core.control import Control
import clickqt.widgets


def findChildren(
    widget: QWidget,
    child_type: QWidget,
    options=Qt.FindChildOption.FindDirectChildrenOnly,
) -> t.Sequence:
    return widget.findChildren(child_type, options=options)


def checkLen(children: t.Sequence, expected_len: int) -> t.Sequence:
    assert len(children) == expected_len
    return children


def hasWidgets(
    tab_widget_content: QWidget,
    control: Control,
    group_hierarchy_name: str,
    params: t.Sequence[click.Parameter],
) -> tuple[bool, str]:
    for param in params:
        # Search for the widget of type 'widget_type' and name 'widget_name' recursively
        children = tab_widget_content.findChildren(
            control.widget_registry[group_hierarchy_name][param.name].widget_type,
            (control.widget_registry[group_hierarchy_name][param.name].widget_name),
        )
        if len(children) == 0:
            return (False, f"Widget is missing in QTabWidget: '{param.name}'")
        if isinstance(
            control.widget_registry[group_hierarchy_name][param.name],
            clickqt.widgets.ConfirmationWidget,
        ):
            if len(children) != 1 + 2:  # Container widget and the two normal widgets
                return (
                    False,
                    f"ConfirmationWidget is multiple times in QTabWidget: '{param.name}'",
                )
        elif len(children) != 1:
            return (False, f"Widget is multiple times in QTabWidget: '{param.name}'")

    return (True, "")


def isIncluded(
    tab_widget: QWidget,
    expected_group_command: t.Sequence[click.Command],
    control: Control,
    group_hierarchy_name: str,
) -> tuple[bool, str]:
    # exact type check needed
    if type(tab_widget) is QWidget:  # Group has options
        tab_widget = checkLen(findChildren(tab_widget, QTabWidget), 1)[0]

    assert tab_widget.count() == len(
        expected_group_command
    ), "Amount of tabs != Amount of commands and groups"

    for group_command in expected_group_command:
        # group_command.name is the name of one tab
        for i in range(tab_widget.count()):
            tab_widget_name = tab_widget.tabText(i)

            if tab_widget_name == group_command.name:
                res = hasWidgets(
                    tab_widget.widget(i),
                    control,
                    control.concat(group_hierarchy_name, group_command.name),
                    group_command.params,
                )
                if not res[0]:
                    return res

                break  # Skips the else block
        else:
            return (
                False,
                f"Command-/Group name is missing in QTabWidget: '{group_command.name}'",
            )

        # Recursive call for groups
        if isinstance(group_command, click.Group):
            res = isIncluded(
                next(
                    filter(
                        lambda x: isinstance(x, QTabWidget) or type(x) is QWidget,
                        [tab_widget.widget(i) for i in range(tab_widget.count())],
                    )
                ),
                group_command.commands.values(),
                control,
                control.concat(group_hierarchy_name, group_command.name),
            )
            if not res[0]:
                return res

    return (True, "")


@pytest.mark.parametrize(
    ("root_group_command"),
    [
        (click.Command("cli", params=[])),
        (click.Group("group", commands=[click.Command("cli", params=[])])),
        (
            click.Group(
                "root_group",
                commands=[
                    click.Group(
                        "sub_group", commands=[click.Command("sub_cli", params=[])]
                    ),
                    click.Command("cli", params=[]),
                ],
            )
        ),
        (
            click.Group(
                "root_group",
                commands=[
                    click.Command("cli1", params=[]),
                    click.Command("cli2", params=[]),
                    click.Group(
                        "sub_group",
                        commands=[
                            click.Command("sub_cli1", params=[]),
                            click.Command("sub_cli2", params=[]),
                            click.Group(
                                "sub_sub_group",
                                commands=[
                                    click.Command("sub_sub_cli1", params=[]),
                                    click.Command("sub_sub_cli2", params=[]),
                                ],
                            ),
                        ],
                    ),
                ],
            )
        ),
    ],
)
def test_gui_construction_no_options(root_group_command: click.Command):
    control = clickqt.qtgui_from_click(root_group_command)
    gui = control.gui

    # Base widgets are set correctly
    assert checkLen(findChildren(gui.window, QSplitter), 1)[0] == gui.splitter
    buttons = checkLen(
        findChildren(
            gui.splitter, QPushButton, Qt.FindChildOption.FindChildrenRecursively
        ),
        3,
    )
    assert (
        gui.run_button in buttons
        and gui.stop_button in buttons
        and gui.copy_button in buttons
    )
    assert (
        checkLen(findChildren(gui.splitter, TerminalOutput), 1)[0]
        == gui.terminal_output
    )

    # Check for right amount of QTabWidgets-instances with correct tab-names
    if (
        isinstance(root_group_command, click.Group)
        and len(root_group_command.commands.values()) > 0
    ):
        parent_tab_widget = checkLen(findChildren(gui.splitter, QTabWidget), 1)[0]
        assert parent_tab_widget == gui.widgets_container

        included, err_message = isIncluded(
            parent_tab_widget,
            root_group_command.commands.values(),
            control,
            root_group_command.name,
        )

        assert included, err_message
    else:
        checkLen(
            [
                x
                for x in findChildren(gui.splitter, QWidget)
                if x == gui.widgets_container
            ],
            1,
        )


@pytest.mark.parametrize(
    ("root_group_command"),
    [
        (
            click.Command(
                "cli",
                params=[
                    click.Option(param_decls=["--test1"], **ClickAttrs.checkbox()),
                    click.Option(param_decls=["--test2"], **ClickAttrs.intfield()),
                ],
            )
        ),
        (
            click.Group(
                "group",
                params=[
                    click.Option(param_decls=["--abc1"], **ClickAttrs.realfield()),
                    click.Option(param_decls=["--abc2"], **ClickAttrs.textfield()),
                ],
                commands=[
                    click.Command(
                        "cli",
                        params=[
                            click.Option(
                                param_decls=["--abc1"], **ClickAttrs.passwordfield()
                            ),
                            click.Option(
                                param_decls=["--abc2"],
                                **ClickAttrs.combobox(choices=["A", "B"]),
                            ),
                        ],
                    )  # Same option names are allowed
                ],
            )
        ),
        (
            click.Group(
                "root_group",
                params=[
                    click.Option(param_decls=["--root1"], **ClickAttrs.datetime()),
                    click.Option(param_decls=["--root2"], **ClickAttrs.uuid()),
                ],
                commands=[
                    click.Group(
                        "sub_group",
                        params=[
                            click.Option(
                                param_decls=["--group1"], **ClickAttrs.intrange()
                            ),
                            click.Option(
                                param_decls=["--group2"], **ClickAttrs.floatrange()
                            ),
                        ],
                        commands=[
                            click.Command(
                                "sub_cli",
                                params=[
                                    click.Option(
                                        param_decls=["--abc1"],
                                        **ClickAttrs.tuple_widget(
                                            types=(click.types.Path(), int)
                                        ),
                                    ),
                                    click.Option(
                                        param_decls=["--abc2"],
                                        **ClickAttrs.multi_value_widget(nargs=2),
                                    ),
                                ],
                            )
                        ],
                    ),
                    click.Command(
                        "cli",
                        params=[
                            click.Option(
                                param_decls=["--abc2"],
                                **ClickAttrs.confirmation_widget(),
                            )
                        ],
                    ),
                ],
            )
        ),
        (
            click.Group(
                "root_group",
                params=[
                    click.Option(param_decls=["--group1"], **ClickAttrs.datetime())
                ],
                commands=[
                    click.Command(
                        "cli1",
                        params=[
                            click.Option(
                                param_decls=["--group1"],
                                **ClickAttrs.multi_value_widget(nargs=2),
                            )
                        ],
                    ),
                    click.Command("cli2", params=[]),
                    click.Group(
                        "sub_group",
                        params=[
                            click.Option(
                                param_decls=["--group1"], **ClickAttrs.filefield()
                            ),
                            click.Option(
                                param_decls=["--group2"], **ClickAttrs.filepathfield()
                            ),
                        ],
                        commands=[
                            click.Command(
                                "sub_cli1",
                                params=[
                                    click.Option(
                                        param_decls=["--cli"],
                                        **ClickAttrs.nvalue_widget(),
                                    )
                                ],
                            ),
                            click.Command("sub_cli2", params=[]),
                            click.Group(
                                "sub_sub_group",
                                params=[
                                    click.Option(
                                        param_decls=["--group1"],
                                        **ClickAttrs.checkbox(),
                                    ),
                                    click.Option(
                                        param_decls=["--group2"],
                                        **ClickAttrs.intrange(),
                                    ),
                                ],
                                commands=[
                                    click.Command("sub_sub_cli1", params=[]),
                                    click.Command(
                                        "sub_sub_cli2",
                                        params=[
                                            click.Option(
                                                param_decls=["--cli2"],
                                                **ClickAttrs.tuple_widget(
                                                    types=(
                                                        click.types.FloatRange(),
                                                        int,
                                                    )
                                                ),
                                            )
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            )
        ),
    ],
)
def test_gui_construction_with_options(root_group_command: click.Command):
    control = clickqt.qtgui_from_click(root_group_command)
    gui = control.gui

    # Check for right amount of QTabWidgets-instances with correct tab-names and correct widget objects
    if (
        isinstance(root_group_command, click.Group)
        and len(root_group_command.commands.values()) > 0
    ):
        parent_tab_widget: QWidget = None

        if len(root_group_command.params) > 0:
            parent_tab_widget = checkLen(
                [
                    x
                    for x in findChildren(
                        gui.splitter,
                        QWidget,
                        Qt.FindChildOption.FindChildrenRecursively,
                    )
                    if x == gui.widgets_container
                ],
                1,
            )[0]
        else:
            parent_tab_widget = checkLen(findChildren(gui.splitter, QTabWidget), 1)[0]
            assert parent_tab_widget == gui.widgets_container

        hasWidgets(
            parent_tab_widget,
            control,
            root_group_command.name,
            root_group_command.params,
        )

        included, err_message = isIncluded(
            parent_tab_widget,
            root_group_command.commands.values(),
            control,
            root_group_command.name,
        )

        assert included, err_message
    else:
        checkLen(
            [
                x
                for x in findChildren(gui.splitter, QWidget)
                if x == gui.widgets_container
            ],
            1,
        )


def test_gui_start_stop_execution():
    param = click.Option(param_decls=["--p"], **ClickAttrs.checkbox())
    cli = click.Command("cli", params=[param], callback=lambda p: QThread.msleep(100))

    control = clickqt.qtgui_from_click(cli)
    run_button = control.gui.run_button
    stop_button = control.gui.stop_button

    assert run_button.isEnabled() and not stop_button.isEnabled()
    assert control.worker is None and control.worker_thread is None

    run_button.click()  # Start execution
    wait_process_Events(1)  # Wait for starting the worker

    assert not run_button.isEnabled() and stop_button.isEnabled()
    assert control.worker is not None and control.worker_thread is not None

    stop_button.click()  # Stop execution
    wait_process_Events(1)  # Wait for stopping the worker

    assert run_button.isEnabled() and not stop_button.isEnabled()
    assert control.worker is None and control.worker_thread is None
    assert "Execution stopped!\n" in control.gui.terminal_output.toPlainText()

    run_button.click()  # Start execution
    wait_process_Events(1)  # Wait for starting the worker

    # Wait for thread to finish
    wait_process_Events(100, 5)
    # """ # QSignalSpy problematic with Python 3.9 (core dumped)
    # spy = QSignalSpy(control.worker, SIGNAL("finished()"))
    # is_finished = False

    # for _ in range(10):
    #     is_finished = spy.count() > 0

    #     QApplication.processEvents()
    #     spy.wait(100)

    #     if is_finished:
    #         break
    # """

    assert run_button.isEnabled() and not stop_button.isEnabled()
    assert control.worker is None and control.worker_thread is None


@pytest.mark.parametrize(
    ("exception", "output_expected"),
    [
        (SystemExit(527), "SystemExit-Exception, return code: 527\n"),
        (TypeError("Wrong type"), "Exception: Wrong type\n"),
    ],
)
def test_gui_exception(exception: Exception, output_expected: str):
    param = click.Option(param_decls=["--p"], **ClickAttrs.checkbox())
    cli = click.Command("cli", params=[param], callback=lambda p: raise_(exception))

    control = clickqt.qtgui_from_click(cli)
    run_button = control.gui.run_button

    run_button.click()  # Start execution
    wait_process_Events(1)  # Wait for starting the worker

    # Worker thread does not sleep so no need to wait for thread to finish

    assert output_expected in control.gui.terminal_output.toPlainText()



from __future__ import annotations

from enum import IntFlag
import typing as t

import click
from PySide6.QtWidgets import QLineEdit, QPushButton, QFileDialog, QHBoxLayout, QWidget
from PySide6.QtCore import QDir

from clickqt.widgets.core.QPathDialog import QPathDialog
from clickqt.widgets.basewidget import BaseWidget
from clickqt.core.utils import remove_prefix

try:
    from enum_tools.documentation import document_enum
except ImportError:  # pragma: no cover
    document_enum = lambda x: x # pylint: disable=unnecessary-lambda-assignment


class TextField(BaseWidget):
    """Represents a click.types.StringParamType-object and user defined click types.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QLineEdit

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        if self.parent_widget is None:
            if (
                envvar_value := param.resolve_envvar_value(click.Context(self.click_command))
            ) is not None:  # Consider envvar
                self.set_value(envvar_value)
            else:  # Consider default value
                self.set_value(BaseWidget.get_param_default(param, ""))

    def set_value(self, value: t.Any):
        if isinstance(value, str):
            self.widget.setText(value)
        else:
            self.widget.setText(
                self.type.convert(
                    value=value,
                    param=self.click_command,
                    ctx=click.Context(self.click_command),
                )
            )

    def is_empty(self) -> bool:
        """Returns True if the current text is an empty string, False otherwise."""

        return self.get_widget_value() == ""

    def get_widget_value(self) -> str:
        return self.widget.text()


class PathField(TextField):
    """Provides basic functionalities for click.types.File- and click.types.Path-objects.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    @document_enum
    class FileType(IntFlag):
        """Specifies the possible file types."""

        Unknown = 0
        File = 1  # doc: The widget accepts files.
        Directory = 2  # doc: The widget accepts directories.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        self.file_type: PathField.FileType = (
            PathField.FileType.Unknown
        )  #: File type of this widget, defaults to :attr:`~clickqt.widgets.textfield.PathField.FileType.Unknown`.

        self.browse_btn = QPushButton("Browse")
        self.browse_btn.clicked.connect(self.browse)
        self.layout.removeWidget(self.widget)
        input_btn_container = QWidget()
        input_btn_container.setLayout(QHBoxLayout())
        input_btn_container.layout().setContentsMargins(0, 0, 0, 0)
        input_btn_container.layout().addWidget(self.widget)
        input_btn_container.layout().addWidget(self.browse_btn)
        self.layout.addWidget(input_btn_container)

    def set_value(self, value: t.Any):
        self.widget.setText(str(value))

    def browse(self):
        """Opens a :class:`~clickqt.widgets.core.QPathDialog.QPathDialog` if :attr:`~clickqt.widgets.textfield.PathField.file_type` is of type
        :attr:`~clickqt.widgets.textfield.PathField.FileType.File` and :attr:`~clickqt.widgets.textfield.PathField.FileType.Directory`, a
        :class:`~PySide6.QtWidgets.QFileDialog` otherwise. Sets the relative path or absolute path (-> path does not contain the path of this project)
        that was selected in the dialog as the value of the widget.
        """

        assert self.file_type != PathField.FileType.Unknown

        if (
            self.file_type & PathField.FileType.File
            and self.file_type & PathField.FileType.Directory
        ):
            dialog = QPathDialog(
                None, directory=QDir.currentPath(), exist=self.type.exists
            )
            if dialog.exec():
                self.handle_valid(True)
                self.set_value(
                    remove_prefix(
                        dialog.selectedPath()
                        .replace(QDir.currentPath(), "")
                        .replace("/", QDir.separator()),
                        QDir.separator(),
                    )
                )
        else:
            dialog = QFileDialog(directory=QDir.currentPath())
            dialog.setViewMode(QFileDialog.ViewMode.Detail)
            dialog.setOption(QFileDialog.Option.DontUseNativeDialog, True)
            # File or directory selectable
            if self.file_type == PathField.FileType.File:
                # click.File hasn't "exists" attribute, click.Path hasn't "mode" attribute
                if (hasattr(self.type, "exists") and self.type.exists) or (
                    hasattr(self.type, "mode") and "r" in self.type.mode
                ):
                    dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
                else:
                    dialog.setFileMode(QFileDialog.FileMode.AnyFile)
            else:  # Only FilePathField can be here
                dialog.setFileMode(QFileDialog.FileMode.Directory)
                dialog.setOption(QFileDialog.Option.ShowDirsOnly, True)

            if dialog.exec():
                filenames = dialog.selectedFiles()
                if filenames and len(filenames) > 0:
                    self.handle_valid(True)
                    self.set_value(
                        remove_prefix(
                            filenames[0]
                            .replace(QDir.currentPath(), "")
                            .replace("/", QDir.separator()),
                            QDir.separator(),
                        )
                    )



from __future__ import annotations

import re


def is_nested_list(lst):
    if isinstance(lst, list):
        return any(isinstance(item, list) for item in lst)
    return False


def is_file_path(string):
    # Regular expression pattern to match file paths
    pattern = r"^([a-zA-Z]:)?[\\/](?:[^\0<>:\/\\|?*\n]+[\\/])*[^\0<>:\/\\|?*\n]*$"

    # Check if the string matches the pattern
    return re.match(pattern, string)


def remove_prefix(text: str, prefix: str):
    if text.startswith(prefix):
        return text[len(prefix) :]

    return text



from __future__ import annotations

import typing as t

import click
import pytest

from tests.testutils import ClickAttrs, raise_
import clickqt.widgets
from clickqt.core.error import ClickQtError


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected"),
    [
        (
            ClickAttrs.checkbox(callback=lambda ctx, param, value: not value),
            True,
            False,
        ),
        (ClickAttrs.intfield(callback=lambda a, b, value: value - 1), 12, 11),
        (ClickAttrs.realfield(callback=lambda a, b, value: value + 1.5), 10.5, 12),
        (ClickAttrs.realfield(callback=lambda a, b, value: None), -312.2, None),
        (
            ClickAttrs.realfield(callback=lambda a, b, value: "test"),
            14.2,
            "test",
        ),  # Return type can be completely different
        (
            ClickAttrs.datetime(callback=lambda a, b, value: "2023-06-23 15:14:20"),
            "2020-01-21 10:11:12",
            "2023-06-23 15:14:20",
        ),
        (
            ClickAttrs.nvalue_widget(type=float, callback=lambda a, b, value: 1),
            [14.2, -2.3],
            1,
        ),
        (
            ClickAttrs.tuple_widget(
                types=(int, str), callback=lambda a, b, value: [value[0] + 5, "test"]
            ),
            [10, "10"],
            [15, "test"],
        ),
        (
            ClickAttrs.multi_value_widget(
                nargs=2,
                type=int,
                callback=lambda a, b, value: [value[0] + 5, value[1] - 5],
            ),
            [10, 10],
            [15, 5],
        ),
        (
            ClickAttrs.nvalue_widget(
                type=(int, (str, float)),
                callback=lambda a, b, value: [
                    [value[0][0] + 4, [value[0][1][0] + "est", value[0][1][1] + 3]]
                ],
            ),
            [[1, ["t", 2.1]]],
            [[5, ["test", 5.1]]],
        ),
    ],
)
def test_callback(click_attrs: dict, value: t.Any, expected: t.Any):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    control.widget_registry[cli.name][param.name].set_value(value)
    val, err = control.widget_registry[cli.name][param.name].get_value()

    assert val == expected and err.type == ClickQtError.ErrorType.NO_ERROR


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected"),
    [
        (
            ClickAttrs.checkbox(
                callback=lambda ctx, param, value: raise_(Exception("..."))
            ),
            False,
            None,
        ),
        (
            ClickAttrs.intfield(
                callback=lambda a, b, c: raise_(click.BadParameter("..."))
            ),
            123,
            None,
        ),
        (
            ClickAttrs.nvalue_widget(
                type=(int, (str, float)),
                callback=lambda a, b, c: raise_(click.BadParameter("...")),
            ),
            [[1, ["t", 2.1]]],
            None,
        ),
    ],
)
def test_callback_fail(click_attrs: dict, value: t.Any, expected: t.Any):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    control.widget_registry[cli.name][param.name].set_value(value)
    val, err = control.widget_registry[cli.name][param.name].get_value()

    assert val == expected and err.type == ClickQtError.ErrorType.PROCESSING_VALUE_ERROR


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected"),
    [
        (ClickAttrs.intfield(callback=lambda ctx, param, value: ctx.abort()), 1, None),
        (
            ClickAttrs.nvalue_widget(
                type=(int, str), callback=lambda ctx, param, value: ctx.abort()
            ),
            [[1, "t"]],
            None,
        ),
    ],
)
def test_callback_abort(click_attrs: dict, value: t.Any, expected: t.Any):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    control.widget_registry[cli.name][param.name].set_value(value)
    val, err = control.widget_registry[cli.name][param.name].get_value()

    assert val == expected and err.type == ClickQtError.ErrorType.ABORTED_ERROR


@pytest.mark.parametrize(
    ("click_attrs", "value", "expected"),
    [
        (
            ClickAttrs.textfield(callback=lambda ctx, param, value: ctx.exit()),
            "test",
            None,
        ),
        (
            ClickAttrs.nvalue_widget(
                type=(int, str), callback=lambda ctx, param, value: ctx.exit()
            ),
            [[1, "t"]],
            None,
        ),
    ],
)
def test_callback_exit(click_attrs: dict, value: t.Any, expected: t.Any):
    param = click.Option(param_decls=["--test"], **click_attrs)
    cli = click.Command("cli", params=[param])

    control = clickqt.qtgui_from_click(cli)
    control.widget_registry[cli.name][param.name].set_value(value)
    val, err = control.widget_registry[cli.name][param.name].get_value()

    assert val == expected and err.type == ClickQtError.ErrorType.EXIT_ERROR



from __future__ import annotations

import typing as t

import click
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QIcon

from clickqt.core.control import Control


def qtgui_from_click(cmd: click.Command, application_name: t.Optional[str] = None, window_icon: t.Optional[str] = None):
    """This function is used to generate the GUI for a given command. It takes a click command as its argument and returns a Control object
    that contains the GUI, execution logic, and the generated widgets used for the parameters of the command.

    :param cmd: The click.Command-object to create a GUI from
    :param application_name: Name of the application, defaults to None (= 'python')
    :param window_icon: Path to an icon, changes the icon of the application, defaults to None (= no icon)
                  
    :return: The control-object that contains the GUI
    """
    if QApplication.instance() is None:
        # Testing: The testing suite creates a QApplication instance
        app = QApplication([])
        app.setWindowIcon(QIcon(window_icon))
        app.setApplicationName(application_name)
        app.setStyleSheet(
            """QToolTip { 
                background-color: #182035; 
                color: white; 
                border: white solid 1px
                }"""
        )

    return Control(cmd)



from __future__ import annotations

import typing as t
import click
from PySide6.QtWidgets import QVBoxLayout, QScrollArea, QPushButton, QWidget
from PySide6.QtCore import Qt

from clickqt.widgets.basewidget import BaseWidget, MultiWidget
from clickqt.core.error import ClickQtError


class NValueWidget(MultiWidget):
    """Represents a multiple click.Parameter-object.
    The child widgets are set according to :func:`~clickqt.widgets.basewidget.MultiWidget.init`.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param widgetsource: A reference to :func:`~clickqt.core.gui.GUI.create_widget`
    :param parent: The parent BaseWidget of **otype**, defaults to None. Needed for :class:`~clickqt.widgets.basewidget.MultiWidget`-widgets
    :param kwargs: Additionally parameters ('com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QScrollArea  #: The Qt-type of this widget.

    def __init__(
        self,
        otype: click.ParamType,
        param: click.Parameter,
        widgetsource: t.Callable[[t.Any], BaseWidget],
        parent: t.Optional[BaseWidget] = None,
        **kwargs,
    ):
        super().__init__(otype, param, parent=parent, **kwargs)

        assert not isinstance(
            otype, click.Choice
        ), f"'otype' is of type '{click.Choice}', but there is a better version for this type"
        assert param.multiple, "'param.multiple' should be True"

        self.widget.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.optkwargs = kwargs
        self.widgetsource = widgetsource
        self.vbox = QWidget()
        self.vbox.setLayout(QVBoxLayout())
        self.widget.setWidgetResizable(True)
        addfieldbtn = QPushButton("+", self.widget)
        # Add an empty widget
        addfieldbtn.clicked.connect(lambda: self.add_pair())  # pylint: disable=unnecessary-lambda
        self.vbox.layout().addWidget(addfieldbtn)
        self.widget.setWidget(self.vbox)
        self.buttondict: dict[QPushButton, BaseWidget] = {}

        self.children = self.buttondict.values()

        self.init()

    def add_pair(self, value: t.Any = None):
        """Adds a new (child-)widget of type **otype** with a remove button to this widget.
        If **value** is not None, it will be the initial value of the new added widget.

        :param value: The initial value of the new widget, defaults to None (=widget will be zero initialized)
        """

        if len(self.children) == 0:
            self.handle_valid(True)

        self.param.multiple = (
            False  # nargs cannot be nested, so it is safe to turn this off for children
        )
        clickqtwidget: BaseWidget = self.widgetsource(
            self.type,
            self.param,
            widgetsource=self.widgetsource,
            parent=self,
            **self.optkwargs,
        )
        self.param.multiple = True  # click needs this for a correct conversion
        if value is not None:
            clickqtwidget.set_value(value)
        clickqtwidget.layout.removeWidget(clickqtwidget.label)
        clickqtwidget.label.deleteLater()
        removebtn = QPushButton("-", clickqtwidget.widget)
        clickqtwidget.layout.addWidget(removebtn)
        removebtn.clicked.connect(lambda: self.remove_button_pair(removebtn))
        self.vbox.layout().addWidget(clickqtwidget.container)
        self.buttondict[removebtn] = clickqtwidget
        self.widget.setWidget(self.vbox)

    def remove_button_pair(self, btn_to_remove: QPushButton):
        """Removes the widget assoziated with **btn_to_remove**.

        :param btn_to_remove: The remove-button that was clicked
        """

        if btn_to_remove in self.buttondict:
            cqtwidget = self.buttondict.pop(btn_to_remove)
            self.vbox.layout().removeWidget(cqtwidget.container)
            cqtwidget.layout.removeWidget(cqtwidget.widget)
            cqtwidget.container.deleteLater()
            btn_to_remove.deleteLater()
            QScrollArea.updateGeometry(self.widget)

    def get_value(self) -> tuple[t.Any, ClickQtError]:
        """Validates the value of the children-widgets and returns the result. If multiple errors occured then they will be concatenated and returned.

        :return: Valid: (children-widget values or the value of a callback, :class:`~clickqt.core.error.ClickQtError.ErrorType.NO_ERROR`)\n
                 Invalid: (None, :class:`~clickqt.core.error.ClickQtError.ErrorType.CONVERTING_ERROR` or
                 :class:`~clickqt.core.error.ClickQtError.ErrorType.PROCESSING_VALUE_ERROR` or :class:`~clickqt.core.error.ClickQtError.ErrorType.REQUIRED_ERROR`)
        """

        value_missing = False
        if len(self.children) == 0:
            default = BaseWidget.get_param_default(self.param, None)

            if self.param.required and default is None:
                self.handle_valid(False)
                return (
                    None,
                    ClickQtError(
                        ClickQtError.ErrorType.REQUIRED_ERROR,
                        self.widget_name,
                        self.param.param_type_name,
                    ),
                )
            if (
                envvar_values := self.param.value_from_envvar(
                    click.Context(self.click_command)
                )
            ) is not None:
                for ev in envvar_values:
                    self.add_pair(ev)
            elif default is not None:  # Add new pairs
                for (
                    value
                ) in (
                    default
                ):  # All defaults will be considered if len(self.children)) == 0
                    self.add_pair(value)
            else:  # param is not required and there is no default -> value is None
                value_missing = True  # But callback should be considered

        values: t.Optional[t.Iterable] = None

        if not value_missing:
            values = []
            err_messages: list[str] = []
            default = BaseWidget.get_param_default(self.param, None)

            # len(self.children)) < len(default): We set at most len(self.children)) defaults
            # len(self.children)) >= len(default): All defaults will be considered
            for i, child in enumerate(self.children):
                if child.is_empty():
                    if child.param.required and default is None:
                        self.handle_valid(False)
                        return (
                            None,
                            ClickQtError(
                                ClickQtError.ErrorType.REQUIRED_ERROR,
                                child.widget_name,
                                child.param.param_type_name,
                            ),
                        )
                    if default is not None and i < len(
                        default
                    ):  # Overwrite the empty widget with the default value (if one exists)
                        child.set_value(
                            default[i]
                        )  # If the widget is a tuple, all values will be overwritten
                    else:  # No default exists -> Don't consider the value of this child
                        # We can't remove the child because there would be a problem with out-of-focus-validation when
                        # having multiple string based widgets (the validator would remove the widget before all child-widgets could be filled)
                        continue

                try:  # Try to convert the provided value into the corresponding click object type
                    values.append(
                        self.type.convert(
                            value=child.get_widget_value(),
                            param=self.param,
                            ctx=click.Context(self.click_command),
                        )
                    )
                    child.handle_valid(True)
                except Exception as e: # pylint: disable=broad-exception-caught
                    child.handle_valid(False)
                    err_messages.append(str(e))

            if len(err_messages) > 0:  # Join all error messages and return them
                messages = ", ".join(err_messages)
                return (
                    None,
                    ClickQtError(
                        ClickQtError.ErrorType.CONVERTING_ERROR,
                        self.widget_name,
                        messages
                        if len(err_messages) == 1
                        else messages.join(["[", "]"]),
                    ),
                )

            if len(values) == 0:  # All widgets are empty
                values = None

        return self.handle_callback(values)

    def set_value(self, value: t.Iterable[t.Any]):
        """Sets the values of the (child-)widgets.
        The number of (child-)widgets are adjusted to the length of **value**. This means that (child-)widgets may be added, but also removed.

        :param value: The list of new values that should be stored in the (child-)widgets
        """

        if len(value) < len(self.children):  # Remove pairs
            for btns in list(self.buttondict.keys())[len(value) :]:
                self.remove_button_pair(btns)
        if len(value) > len(self.children):  # Add pairs
            for i in range(len(value) - len(self.children)):
                self.add_pair()

        for i, c in enumerate(self.children):  # Set the value
            c.set_value(value[i])

    def handle_valid(self, valid: bool):
        if len(self.children) == 0:
            BaseWidget.handle_valid(self, valid)
        else:
            super().handle_valid(valid)



from __future__ import annotations

import datetime
import typing as t

import click
from PySide6.QtWidgets import QDateTimeEdit
from PySide6.QtCore import QDateTime, Qt
from PySide6.QtGui import QAction, QActionGroup

from clickqt.widgets.basewidget import BaseWidget


class DateTimeEdit(BaseWidget):
    """Represents a click.types.DateTime object. The format of the datetime displayed in this widget can be changed by opening the context menu.

    :param otype: The type which specifies the clickqt widget type. This type may be different compared to **param**.type when dealing with click.types.CompositeParamType-objects
    :param param: The parameter from which **otype** came from
    :param kwargs: Additionally parameters ('parent', 'widgetsource', 'com', 'label') needed for
                    :class:`~clickqt.widgets.basewidget.MultiWidget`- / :class:`~clickqt.widgets.confirmationwidget.ConfirmationWidget`-widgets
    """

    widget_type = QDateTimeEdit  #: The Qt-type of this widget.

    def __init__(self, otype: click.ParamType, param: click.Parameter, **kwargs):
        super().__init__(otype, param, **kwargs)

        assert isinstance(
            otype, click.DateTime
        ), f"'otype' must be of type '{click.DateTime}', but is '{type(otype)}'."

        self.widget.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)

        #: Contains all available formats in qt_format- and click_format-style. Default format string is the last string in the *click.types.DateTime*\s ``formats`` attribute
        self.format_group: QActionGroup = QActionGroup(self.widget)

        # Add for each format a context menu entry
        for (
            click_format
        ) in self.type.formats:  # click ensures that there is at least one format
            qt_format: str = (
                click_format.replace("%Y", "yyyy")
                .replace("%m", "MM")
                .replace("%d", "dd")
                .replace("%H", "hh")
                .replace("%M", "mm")
                .replace("%S", "ss")
            )
            action = QAction(qt_format, self.widget)
            action.setData(click_format)
            action.setCheckable(True)
            if (
                click_format == self.type.formats[-1]
            ):  # Last format is checked at the beginning
                action.setChecked(True)
                self.widget.setDisplayFormat(qt_format)
            else:
                action.setChecked(False)
            self.widget.addAction(action)
            self.format_group.addAction(action)

        self.format_group.setExclusive(
            True
        )  # Only one format can be checked at the same time
        self.format_group.triggered.connect(
            lambda action: self.widget.setDisplayFormat(action.text())
            or self.container.setFocus()
        )  # Update display format and geometry

        if (
            self.parent_widget is None
            and (default := BaseWidget.get_param_default(param, None)) is not None
        ):
            self.set_value(default)

    def set_value(self, value: t.Any):
        """Sets the value of the Qt-widget according to the selected format stored in :attr:`~clickqt.widgets.datetimeedit.DateTimeEdit.format_group`."""
        # value -> datetime -> str -> QDateTime
        self.widget.setDateTime(
            QDateTime.fromString(
                self.type.convert(
                    str(value), self.click_command, click.Context(self.click_command)
                ).strftime(self.format_group.checkedAction().data()),
                self.format_group.checkedAction().text(),
            )
        )

    def get_widget_value(self) -> datetime.datetime:
        return self.widget.dateTime().toPython()
